open Format
open Formatx
open Term

type t = Rule.t list


let variables rules =
  Listx.unique [ x | rule <- rules; x <- Rule.variables rule ]

let functions rules =
  Listx.unique [ x | rule <- rules; x <- Rule.functions rule ]

let signature rules =
  Listx.unique [ x | rule <- rules; x <- Rule.signature rule ]

let defined_symbols rules =
  Listx.unique [ Term.root l | l, _ <- rules ]

let is_non_root_overlapping s t = not (Subst.unifiable (Term.rename s) t)

let is_non_overlapping rules =
  List.for_all (fun (l1, s1, r1) ->
    List.for_all (fun (l2, r2) -> 
      ((l1, r1) = (l2, r2) && l1 = s1) ||
      is_non_root_overlapping s1 l2) rules)
    [ l1, s1, r1 | l1, r1 <- rules; s1 <- Term.subterms l1;
      not (Term.is_variable s1) ]

let is_non_duplicating rules = List.for_all Rule.is_non_duplicating rules

let variable_condition rules =
  List.for_all Rule.variable_condition rules

let is_constructor_system rules =
  let ds = defined_symbols rules in
  let p = function
    | Term.V _, _ -> true
    | Term.F (_, ts), _ -> 
        List.for_all (fun t -> Listset.inter (Term.functions t) ds = []) ts in
  List.for_all p rules

let linear rules = List.for_all Rule.linear rules

let left_linear rules = List.for_all Rule.left_linear rules

let right_linear rules = List.for_all Rule.right_linear rules

(* pretty printers *)

let print ppf rules =
  fprintf ppf "@[<v 0>%a@]" (print_list Rule.print "@ ") rules

let print_wst ppf rs =
  fprintf ppf "(VAR %a)@.(RULES@.%a@.)@." 
    (Formatx.print_list pp_print_string " ") (variables rs)
    print rs

let print_std rules = 
 printf "RULES: @.%a@.@." print rules

(* some commonly used printers *)

let print_es_std rs =
 let print_eq ppf (l,r) = 
   fprintf ppf "@[<2>%a =@ %a@]" Term.print l Term.print r
 in
  let print_eqs1 ppf1 rules =
   fprintf ppf1 "@[<v 0>%a@]" (Formatx.print_list print_eq "@ ") rules
 in
  printf "EQUATIONS: @.%a@.@." print_eqs1 rs


let rpl_spcl_char rules = 
 let rec rpl_trm = function
  | V x -> V x
  | F (f,fs) -> (
     match f with
     | "*" -> F ("times", [ rpl_trm fi | fi <- fs])
     | "+" -> F ("plus", [ rpl_trm fi | fi <- fs])
     | "@" -> F ("app", [ rpl_trm fi | fi <- fs]) 
     | "." -> F ("cons", [ rpl_trm fi | fi <- fs])
     | "-" -> F ("minus", [ rpl_trm fi | fi <- fs])
     | "div" ->  F ("divi", [ rpl_trm fi | fi <- fs])
     | "/" -> F ("slash", [ rpl_trm fi | fi <- fs])
     | "false" ->  F ("false_", [ rpl_trm fi | fi <- fs])
     | "true" ->  F ("true_", [ rpl_trm fi | fi <- fs])
     | "or" ->  F ("or_", [ rpl_trm fi | fi <- fs])
     | "and" ->  F ("and_", [ rpl_trm fi | fi <- fs])
     | "not" -> F("not_", [ rpl_trm fi | fi <- fs])
     | _ -> F (f, [ rpl_trm fi | fi <- fs])
 )
 in
 [ rpl_trm s, rpl_trm t | s,t <- rules ]

let rules_over es =
 let rev es1 = [ t,s | s,t <- es1 ] in
 [ rule | rule <- es @ rev es; Rule.is_rule rule ]


